/*      P.cpp
        Playing with C++

        This game program teaches you to type using all fingers

        Program uses the data file  P.dat

        containing a list of keys. Blanks in file are ignored.
         Every new line contains key set for the next level.
         Duplicate keys in line means proportionally
        increased probability for such key.

        To pause game press Esc key.
        To exit game press Esc key once more.
*/

#include "P.h"                  // Data and functions

//------------------------------------------------------------
// main() - program entry point
//------------------------------------------------------------

void main() 
{
  // Play until the data file has been read, or the player has quit
  while (K.Input ()) 
    Play ();
}

//------------------------------------------------------------
// Play() - game loop. Accepts input from player until maximum
// number of characters has been entered, or Esc has been
// pressed.
//------------------------------------------------------------

void Play() 
{
  int     Score   = 0;            // Game score
  int     Key     = 0;            // Last key pressed
  int     KeyNum  = 0;            // Number of keys pressed
  int     ErrNum  = 0;            // Number of error keys pressed

  DrawPicture ();                 // Draw static picture
  Pause ();                       // Press any key to start
  delay (1000);                   // Time to prepare (msec)

  while (KeyNum++ < par::MaxKey)   // Main loop MaxKey times
  {
    K.Get ();                     // Wait for random character to start
    K.Clear ();                   // Clear keyboard buffer
    Sound (BeepAtt);              // Beep to alert user
    K.Draw ();                    // Draw random character
    Score += K.Trace (Key=0);     // Display trace, get Key

    if (Key == Esc)               // Pause when Esc is pressed
      if (Pause()==Esc) 
        exit(0);                  // Exit program if it is pressed again

    if (Key != K.Char)            // User didnt hit correct key...
    {
      Sound (BeepErr);            // Error beep
      DisplayErr (++ErrNum);      // Display error count
    }

    DisplayScore (Score);         // Display the users score
    DisplayBar   (KeyNum);        // Display progress bar
  }
  Pause();
}

//------------------------------------------------------------
// key::Input() - gets the next character to be displayed from
// the data file. Returns 1 if successful, or 0 if EOF has been
// reached.
//------------------------------------------------------------

int key::Input () 
{
  int i = 0;
  while (1) 
  {
    int   Char;
    Char = fgetc (pFILE);           // Get char from file
    if ( Char == EOF ) return 0;    // Return on end of file
    if ( Char == '\n') break;       // Exit on end of LINE
    if ( Char == ' ') continue;     // Ignore blank
    Set [i++] = (char) Char;        // Initialize key::Set
  }
  Len = i;                          // Initialize key::Len

  return 1;
}

//------------------------------------------------------------
// key::key() - constructor for the 'key' class. This routine
// performs the following actions:
// (a) opens the data file, exiting if it can't
// (b) calls Init to initialise the graphics library
// (c) sets the screen font
//------------------------------------------------------------

key::key () 
{
  pFILE=fopen ("P.dat","rt");      // Open data file
  if ( ! pFILE ) exit (1);         // Exit on error

  if (! Init ())                   // Initialize graphics system
   exit (2);                       // Exit on error

  randomize ();                    // For random start char

  // Set text font, size and justification

  int TextSize = 4;
  settextstyle (TRIPLEX_FONT, HORIZ_DIR, TextSize);
  settextjustify (CENTER_TEXT,BOTTOM_TEXT);
}

//------------------------------------------------------------
// key::Init() - initialises the graphics library, loading the
// Borland graphics screen driver, and setting the graphics
// mode. It returns 1 if successful, and 0 on error.
//------------------------------------------------------------

int key::Init() 
{
  int     Driver=DETECT;  // Driver to be used (autodetect it)
  int     Mode;           // Screen mode (returned)
  char *  Path = "";      // Path to driver (current directory)

  initgraph (&Driver, &Mode, Path);

  int InitRes = graphresult();
  if (InitRes != grOk) 
  {
    cout  << "Graphics Error: " << grapherrormsg (InitRes);
    return 0;
  }

  return 1;
}

//------------------------------------------------------------
// key::~key() - destructor for the 'key' class.  This routine
// tidies up by closing the data file, telling the user that the
// game is over, and closing the graphics system.
//------------------------------------------------------------

key::~key ()              // Destructor
{
  fclose (pFILE);         // Close data file
  GameOver ();            // "Game over" message
  closegraph ();          // Close graphics system
}

//------------------------------------------------------------
// DrawPicture() draws the screen display - mountains, snow,
// text, etc.
//------------------------------------------------------------

void DrawPicture() 
{
  int MaxX = getmaxx(),
      MaxY = getmaxy();

  //      Clear screen

  setfillstyle (SOLID_FILL, SkyColor);
  bar (0,0, MaxX,MaxY);

  //      Draw mountain

  int Shape [] = {
        0,16, 3,14, 6,10,
        8,6,                    // Left peak
        10,7,
        12,4,                   // Right peak
        16,12, 20,16
  };
  int NPoints = sizeof Shape / sizeof (int) / 2;
  int i,j;

  for (i=j=0; i<NPoints; i++,j+=2) 
  {
    Shape [j]   = Shape [j]   * getmaxx() / XStep;
    Shape [j+1] = Shape [j+1] * getmaxy() / YStep;
  }

  setcolor (SnowColor);
  setfillstyle (SOLID_FILL, SnowColor);
  fillpoly (NPoints,Shape);       // Draw and fill polygon

  // Draw bottom bar

  setfillstyle (SOLID_FILL, GroundColor);
  bar (Shape[0],Shape[1],MaxX,MaxY);

  // Display score text

  setcolor (TextColor);
  outtextxy (ScoreText.X,ScoreText.Y,"Score ");

  // Display Error text

  outtextxy (ErrorText.X,ErrorText.Y,"Error ");

  DisplayScore (0);
  DisplayErr (0);

  // Display bar

  setfillstyle (SOLID_FILL,BarColor);
  bar (TopBar.X, TopBar.Y, BottomBar.X, BottomBar.Y-1);
}

//------------------------------------------------------------
void key::Get () 
{
  Char = toupper (Set [random (Len)]);
  WhatHand ();
}

//------------------------------------------------------------
// key::WhatHand() determines which hand, right or left, should
// be used to press a given key.
//------------------------------------------------------------

void key::WhatHand () 
{
  static char LeftKeys [] =
        "123456" "!@#$%^"
        "QWERT"
        "ASDFG"
        "ZXCVB";
  int i;

  // Is the character a left-hand key?
  for (i=0; i < sizeof LeftKeys; i++)
    if (LeftKeys [i] == Char) 
    {
      Hand = Left;
      return;
    }

  // If not, it must be a right-hand one.
  Hand = Right;
  return;
}

//------------------------------------------------------------
// key::Clear() - clear the keyboard buffer.
//------------------------------------------------------------

void key::Clear () 
{
  while (kbhit ())
    getch ();
}

//------------------------------------------------------------
// Sound() - sound a tone
//------------------------------------------------------------

void Sound (beep Tone) 
{
  int Freq=0;                 // Frequency, Hz
  int Delay=0;                // Time (milliseconds)

  const int AttFreq = 110;  // Hz
  const int ErrFreq = 440;  // Hz
  const int AttDelay= 50;   // msec
  const int ErrDelay= 200;  // msec

  switch (Tone) 
  {
   case BeepAtt: 
     Freq=AttFreq; 
     Delay=AttDelay; 
     break;
   case BeepErr: 
     Freq=ErrFreq; 
     Delay=ErrDelay; 
     break;
  }
  sound (Freq);
  delay (Delay);
  nosound ();
  delay (Delay);
}

//------------------------------------------------------------
// key::Draw() - output the character that the user has to
// press, above the right or left mountain.
//------------------------------------------------------------

void key::Draw () 
{
  static int X = 0;            // Declared static in order to
  static int Y = 0;            // previous values
  static char Buf [2] = "";

  // Clear previous character
  setcolor(SkyColor); outtextxy (X,Y,Buf);

  // Set position for left or right hand character
  if (Hand == Left) 
  {
    X=LeftChar.X;
    Y = LeftChar.Y;
  }
  else 
  {
    X=RightChar.X;
    Y = RightChar.Y;
  }

  Buf[0] = Char;          // Place new char into Buf

  // Output new char
  setcolor(CharColor); outtextxy (X,Y,Buf);
}

//------------------------------------------------------------
// key::Trace() - displays the 'ski trace' from the top of the
// appropriate mountain, until the user presses a key, or the
// trace reaches the bottom. The routine returns the amount to
// be added to the user's score.
//------------------------------------------------------------

int key::Trace (int & Key) 
{
  // Display ski trace, get key pressed, count score

  int X, Y,                   // Current trace coordinates
      X0, Y0,                 // Trace start
      Finish,                 // Y coordinate of finish
      Amp,Amp2,Amp3;          // Trace amplitudes
                              // Use global constants
  Amp  = MinAmp + random (MaxAmp - MinAmp);
  Amp2 = Amp/4;
  Amp3 = Amp/2;

  static const double PI = asin (1.) * 2.;
  double Freq, Freq2, Freq3;     // Trace frequency

  // Set start and finish points

  if (Hand == Left) 
  {
    X0     = LeftStart.X;
    Y0     = LeftStart.Y;
    Finish = LeftFinish.Y;
  }
  else 
  {
    X0     = RightStart.X;
    Y0     = RightStart.Y;
    Finish = RightFinish.Y;
  }

  // Set frequencies for sin()

  Freq  = 2*PI/(Finish - Y0);
  Freq2 = Freq * 3;
  Freq3 = Freq * 11;

  clock_t StartTime = clock ();   // Get start time

  // Calculate and draw ski trace

  for (X = X0, Y = Y0; Y < Finish; Y++) 
  {
    int   C = TraceColor;

    X = X0 + Amp  * sin (Freq *(Y-Y0))
           + Amp2 * sin (Freq2*(Y-Y0))
           + Amp3 * sin (Freq3*(Y-Y0));

    putpixel (X,  Y,C);
    putpixel (X-1,Y,C);
    putpixel (X+1,Y,C);
    delay (par::Delay);
    if ( kbhit () )               // Test for key pressed
    {
      Key = toupper (getch ());   // Get key from keyboard
      if ( ! Key ) getch ();      // Not ASCII char
      break;                      // Exit on key pressed
    }
  }
  clock_t KeyTime = clock ();     // Set key pressed time
  
  // clock() returns the time in ticks, so to get it in seconds, we divide
  // by the number of ticks per second, which is defined as a constant in time.h
  float Time = (KeyTime - StartTime) / CLK_TCK;

  int Add;                    // Add to score
  if (Y == Finish)
    Add = 50;  // Key not pressed
  else if (Key != Char)
    Add = 10;  // Err key pressed
  else if (Time < 1.0)
    Add =100;  // Time < 1 sec
  else if (Time < 1.5)
    Add = 50;
  else if (Time < 2.0)
    Add = 20;
  else
    Add = 10;

  //      Clear ski trace

  for (X = X0, Y = Y0; Y < Finish; Y++) 
  {
    int C = SnowColor;

    X = X0 + Amp  * sin (Freq *(Y-Y0))
           + Amp2 * sin (Freq2*(Y-Y0))
           + Amp3 * sin (Freq3*(Y-Y0));

    putpixel (X,  Y,C);
    putpixel (X-1,Y,C);
    putpixel (X+1,Y,C);
  }

  return Add;                     // Add to Score
}

//------------------------------------------------------------
// Pause() - wait for a key to be pressed, and return its
// value.
//------------------------------------------------------------

int Pause () 
{
  int Key = getch ();
  if (! Key)
    getch ();

  return Key;
}

//------------------------------------------------------------
// DisplayScore() - display the score value
//------------------------------------------------------------

void DisplayScore (int Score) 
{
  static char Buf[10] = "";

  // Clear previous value
  setcolor (GroundColor);
  outtextxy (ScoreText.X+textwidth("Score "),ScoreText.Y,Buf);

  sprintf (Buf, "%d", Score > 0 ? Score : 0);
  setcolor (ValueColor);
  outtextxy (ScoreText.X+textwidth("Score "),ScoreText.Y,Buf);
}

//------------------------------------------------------------
// DisplayErr() - display the number of errors the user has
// made
//------------------------------------------------------------

void DisplayErr (int ErrNum) 
{
  // Display number of errors
  static  char    Buf [10] = "";

  // Clear previous value
  setcolor (GroundColor);
  outtextxy (ErrorText.X+textwidth("Error "),ErrorText.Y,Buf);

  sprintf (Buf, "%2.0f%%", 100.*ErrNum/par::MaxKey);
  setcolor (ValueColor);
  outtextxy (ErrorText.X+textwidth("Error "),ErrorText.Y,Buf);
}

//------------------------------------------------------------
// DisplayBar() - display the progress bar
//------------------------------------------------------------
void DisplayBar (int N) 
{
  // Decrease bar
  int Y = TopBar.Y + (BottomBar.Y - TopBar.Y) * N/par::MaxKey;

  setfillstyle (SOLID_FILL, SnowColor);
  bar (TopBar.X, TopBar.Y, BottomBar.X, Y-1);
}

//------------------------------------------------------------
// GameOver() - display the 'Game Over' message
//------------------------------------------------------------
void GameOver () 
{
  int TextSize = 8;
  settextstyle (TRIPLEX_FONT, HORIZ_DIR, TextSize);
  settextjustify (CENTER_TEXT,CENTER_TEXT);
  setcolor(BLACK);
  outtextxy (getmaxx()/2,getmaxy()/2,"Game over ");
  Pause();
}

//------------------------------------------------------------
// point:point() - constructor for a 'point' object
//------------------------------------------------------------
point::point (float x,float y) 
{
  // Constructor
  // x and y are relative coordinates

  X = x * getmaxx() / XStep;
  Y = y * getmaxy() / YStep;
}